@@ -5,10 +5,11 @@ module Test.QuickCheck.Arbitrary
   , CoArbitrary(..)
   
   -- ** Helper functions for implementing arbitrary
-  , arbitrarySizedIntegral   -- :: Num a => Gen a
-  , arbitrarySizedFractional -- :: Fractional a => Gen a
-  , arbitraryBoundedIntegral -- :: (Bounded a, Integral a) => Gen a
-  , arbitraryBoundedRandom   -- :: (Bounded a, Random a) => Gen a
+  , arbitrarySizedIntegral        -- :: Num a => Gen a
+  , arbitraryBoundedIntegral      -- :: (Bounded a, Integral a) => Gen a
+  , arbitrarySizedBoundedIntegral -- :: (Bounded a, Integral a) => Gen a
+  , arbitrarySizedFractional      -- :: Fractional a => Gen a
+  , arbitraryBoundedRandom        -- :: (Bounded a, Random a) => Gen a
   -- ** Helper functions for implementing shrink
   , shrinkNothing            -- :: a -> [a]
   , shrinkList               -- :: (a -> [a]) -> [a] -> [[a]]
@@ -56,6 +57,9 @@ import Data.Ratio
   , denominator
   )
 
+import Data.Complex
+  ( Complex((:+)) )
+
 import System.Random
   ( Random
   )
@@ -73,6 +77,9 @@ import Control.Monad
   , liftM5
   )
 
+import Data.Int(Int8, Int16, Int32, Int64)
+import Data.Word(Word, Word8, Word16, Word32, Word64)
+
 --------------------------------------------------------------------------
 -- ** class Arbitrary
 
@@ -97,6 +104,8 @@ instance Arbitrary () where
 
 instance Arbitrary Bool where
   arbitrary = choose (False,True)
+  shrink True = [False]
+  shrink False = []
 
 instance Arbitrary a => Arbitrary (Maybe a) where
   arbitrary = frequency [(1, return Nothing), (3, liftM Just arbitrary)]
@@ -118,31 +127,23 @@ instance Arbitrary a => Arbitrary [a] where
   shrink xs = shrinkList shrink xs
 
 shrinkList :: (a -> [a]) -> [a] -> [[a]]
-shrinkList shr xs = removeChunks xs ++ shrinkOne xs
+shrinkList shr xs = concat [ removes k n xs | k <- takeWhile (>0) (iterate (`div`2) n) ]
+                 ++ shrinkOne xs
  where
-  removeChunks xs = rem (length xs) xs
-   where
-    rem 0 _  = []
-    rem 1 _  = [[]]
-    rem n xs = xs1
-             : xs2
-             : ( [ xs1' ++ xs2 | xs1' <- rem n1 xs1, not (null xs1') ]
-           `ilv` [ xs1 ++ xs2' | xs2' <- rem n2 xs2, not (null xs2') ]
-               )
-     where
-      n1  = n `div` 2
-      xs1 = take n1 xs
-      n2  = n - n1
-      xs2 = drop n1 xs
-
-      []     `ilv` ys     = ys
-      xs     `ilv` []     = xs
-      (x:xs) `ilv` (y:ys) = x : y : (xs `ilv` ys)
+  n = length xs
 
   shrinkOne []     = []
   shrinkOne (x:xs) = [ x':xs | x'  <- shr x ]
                   ++ [ x:xs' | xs' <- shrinkOne xs ] 
 
+  removes k n xs
+    | k > n     = []
+    | null xs2  = [[]]
+    | otherwise = xs2 : map (xs1 ++) (removes k (n-k) xs2)
+   where
+    xs1 = take k xs
+    xs2 = drop k xs
+
 {-
   -- "standard" definition for lists:
   shrink []     = []
@@ -155,6 +156,11 @@ instance (Integral a, Arbitrary a) => Arbitrary (Ratio a) where
   arbitrary = arbitrarySizedFractional
   shrink    = shrinkRealFrac
 
+instance (RealFloat a, Arbitrary a) => Arbitrary (Complex a) where
+  arbitrary = liftM2 (:+) arbitrary arbitrary
+  shrink (x :+ y) = [ x' :+ y | x' <- shrink x ] ++
+                    [ x :+ y' | y' <- shrink y ]
+
 instance (Arbitrary a, Arbitrary b)
       => Arbitrary (a,b)
  where
@@ -200,8 +206,43 @@ instance Arbitrary Integer where
   shrink    = shrinkIntegral
 
 instance Arbitrary Int where
-  --arbitrary = arbitrarySizedIntegral
-  arbitrary = arbitrarySizedBoundedInt
+  arbitrary = arbitrarySizedBoundedIntegral
+  shrink    = shrinkIntegral
+
+instance Arbitrary Int8 where
+  arbitrary = arbitrarySizedBoundedIntegral
+  shrink    = shrinkIntegral
+
+instance Arbitrary Int16 where
+  arbitrary = arbitrarySizedBoundedIntegral
+  shrink    = shrinkIntegral
+
+instance Arbitrary Int32 where
+  arbitrary = arbitrarySizedBoundedIntegral
+  shrink    = shrinkIntegral
+
+instance Arbitrary Int64 where
+  arbitrary = arbitrarySizedBoundedIntegral
+  shrink    = shrinkIntegral
+
+instance Arbitrary Word where
+  arbitrary = arbitrarySizedBoundedIntegral
+  shrink    = shrinkIntegral
+
+instance Arbitrary Word8 where
+  arbitrary = arbitrarySizedBoundedIntegral
+  shrink    = shrinkIntegral
+
+instance Arbitrary Word16 where
+  arbitrary = arbitrarySizedBoundedIntegral
+  shrink    = shrinkIntegral
+
+instance Arbitrary Word32 where
+  arbitrary = arbitrarySizedBoundedIntegral
+  shrink    = shrinkIntegral
+
+instance Arbitrary Word64 where
+  arbitrary = arbitrarySizedBoundedIntegral
   shrink    = shrinkIntegral
 
 instance Arbitrary Char where
@@ -252,8 +293,9 @@ arbitrarySizedFractional =
  where
   precision = 9999999999999 :: Integer
 
--- | Generates an integral number. The number is chosen from the entire
--- range of the type.
+-- | Generates an integral number. The number is chosen uniformly from
+-- the entire range of the type. You may want to use
+-- 'arbitrarySizedBoundedIntegral' instead.
 arbitraryBoundedIntegral :: (Bounded a, Integral a) => Gen a
 arbitraryBoundedIntegral =
   do let mn = minBound
@@ -266,14 +308,18 @@ arbitraryBoundedIntegral =
 arbitraryBoundedRandom :: (Bounded a, Random a) => Gen a
 arbitraryBoundedRandom = choose (minBound,maxBound)
 
--- | Generates an integral number from a bounded domain.
--- Inspired by demands from Phil Wadler.
-arbitrarySizedBoundedInt :: Gen Int
-arbitrarySizedBoundedInt =
+-- | Generates an integral number from a bounded domain. The number is
+-- chosen from the entire range of the type, but small numbers are
+-- generated more often than big numbers. Inspired by demands from
+-- Phil Wadler.
+arbitrarySizedBoundedIntegral :: (Bounded a, Integral a) => Gen a
+arbitrarySizedBoundedIntegral =
   sized $ \s ->
     do let mn = minBound
            mx = maxBound `asTypeOf` mn
-           k  = 2^(s*2 `div` 5)
+           bits n | n `quot` 2 == 0 = 0
+                  | otherwise = 1 + bits (n `quot` 2)
+           k  = 2^(s*(bits mn `max` bits mx `max` 40) `div` 100)
        n <- choose (toInteger mn `max` (-k), toInteger mx `min` k)
        return (fromInteger n `asTypeOf` mn)
 
@@ -288,13 +334,18 @@ shrinkIntegral :: Integral a => a -> [a]
 shrinkIntegral x = 
   nub $
   [ -x
-  | x < 0
+  | x < 0, -x > x
   ] ++
   [ x'
   | x' <- takeWhile (<< x) (0:[ x - i | i <- tail (iterate (`quot` 2) x) ])
   ]
  where
-  x << y = abs x < abs y
+   -- a << b is "morally" abs a < abs b, but taking care of overflow.
+   a << b = case (a >= 0, b >= 0) of
+            (True,  True)  -> a < b
+            (False, False) -> a > b
+            (True,  False) -> a + b < 0
+            (False, True)  -> a + b > 0
 
 -- | Shrink a fraction.
 shrinkRealFrac :: RealFrac a => a -> [a]
@@ -308,7 +359,7 @@ shrinkRealFrac x =
   , x' << x
   ]
  where
-  x << y = abs x < abs y
+  a << b = abs a < abs b
 
 --------------------------------------------------------------------------
 -- ** CoArbitrary
@@ -368,6 +419,9 @@ instance CoArbitrary a => CoArbitrary [a] where
 instance (Integral a, CoArbitrary a) => CoArbitrary (Ratio a) where
   coarbitrary r = coarbitrary (numerator r,denominator r)
 
+instance (RealFloat a, CoArbitrary a) => CoArbitrary (Complex a) where
+  coarbitrary (x :+ y) = coarbitrary x >< coarbitrary y
+
 instance (CoArbitrary a, CoArbitrary b)
       => CoArbitrary (a,b)
  where
@@ -406,6 +460,33 @@ instance CoArbitrary Integer where
 instance CoArbitrary Int where
   coarbitrary = coarbitraryIntegral
 
+instance CoArbitrary Int8 where
+  coarbitrary = coarbitraryIntegral
+
+instance CoArbitrary Int16 where
+  coarbitrary = coarbitraryIntegral
+
+instance CoArbitrary Int32 where
+  coarbitrary = coarbitraryIntegral
+
+instance CoArbitrary Int64 where
+  coarbitrary = coarbitraryIntegral
+
+instance CoArbitrary Word where
+  coarbitrary = coarbitraryIntegral
+
+instance CoArbitrary Word8 where
+  coarbitrary = coarbitraryIntegral
+
+instance CoArbitrary Word16 where
+  coarbitrary = coarbitraryIntegral
+
+instance CoArbitrary Word32 where
+  coarbitrary = coarbitraryIntegral
+
+instance CoArbitrary Word64 where
+  coarbitrary = coarbitraryIntegral
+
 instance CoArbitrary Char where
   coarbitrary = coarbitrary . ord
 
