@@ -7,9 +7,13 @@ module Test.QuickCheck.Exception where
 
 #if defined(__GLASGOW_HASKELL__) && __GLASGOW_HASKELL__ >= 607
 #define GHC_INTERRUPT
+
+#if __GLASGOW_HASKELL__ < 613
+#define GHCI_INTERRUPTED_EXCEPTION
+#endif
 #endif
 
-#if defined OLD_EXCEPTIONS
+#if defined(OLD_EXCEPTIONS)
 import Control.Exception(evaluate, try, Exception(..))
 #else
 import Control.Exception.Extensible(evaluate, try, SomeException(SomeException)
@@ -20,7 +24,9 @@ import Control.Exception.Extensible(evaluate, try, SomeException(SomeException)
 #endif
 
 #if defined(GHC_INTERRUPT)
+#if defined(GHCI_INTERRUPTED_EXCEPTION)
 import Panic(GhcException(Interrupted))
+#endif
 import Data.Typeable
 #if defined(OLD_EXCEPTIONS)
 import Data.Dynamic
@@ -51,9 +57,11 @@ isInterrupt :: AnException -> Bool
 #if defined(OLD_EXCEPTIONS)
 isInterrupt (DynException e) = fromDynamic e == Just Interrupted
 isInterrupt _ = False
-#else
+#elif defined(GHCI_INTERRUPTED_EXCEPTION)
 isInterrupt (SomeException e) =
   cast e == Just Interrupted || cast e == Just UserInterrupt
+#else
+isInterrupt (SomeException e) = cast e == Just UserInterrupt
 #endif
 
 #else /* !defined(GHC_INTERRUPT) */
