@@ -9,9 +9,10 @@ import qualified Test.QuickCheck.Property as P
 import Test.QuickCheck.Text
 import Test.QuickCheck.State
 import Test.QuickCheck.Exception
+import Data.IORef
 
 import System.Random
-  ( RandomGen(..)
+  ( split
   , newStdGen
   , StdGen
   )
@@ -26,7 +27,6 @@ import Data.List
   , groupBy
   , intersperse
   )
-
 --------------------------------------------------------------------------
 -- quickCheck
 
@@ -39,26 +39,35 @@ data Args
   , maxSuccess :: Int                -- ^ maximum number of successful tests before succeeding
   , maxDiscard :: Int                -- ^ maximum number of discarded tests before giving up
   , maxSize    :: Int                -- ^ size to use for the biggest test cases
+  , chatty     :: Bool               -- ^ whether to print anything
   }
  deriving ( Show, Read )
 
 -- | Result represents the test result
 data Result
-  = Success                         -- a successful test run
-    { labels      :: [(String,Int)] -- ^ labels and frequencies found during all tests
+  = Success                            -- a successful test run
+    { numTests       :: Int            -- ^ number of successful tests performed
+    , labels         :: [(String,Int)] -- ^ labels and frequencies found during all tests
+    , output         :: String         -- ^ printed output
     }
-  | GaveUp                          -- given up
-    { numTests    :: Int            -- ^ number of successful tests performed
-    , labels      :: [(String,Int)] -- ^ labels and frequencies found during all tests
+  | GaveUp                             -- given up
+    { numTests       :: Int            -- ^ number of successful tests performed
+    , labels         :: [(String,Int)] -- ^ labels and frequencies found during all tests
+    , output         :: String         -- ^ printed output
     }
-  | Failure                         -- failed test run
-    { usedSeed    :: StdGen         -- ^ what seed was used
-    , usedSize    :: Int            -- ^ what was the test size
-    , reason      :: String         -- ^ what was the reason
-    , labels      :: [(String,Int)] -- ^ labels and frequencies found during all successful tests
+  | Failure                            -- failed test run
+    { numTests       :: Int            -- ^ number of tests performed
+    , numShrinks     :: Int            -- ^ number of successful shrinking steps performed
+    , usedSeed       :: StdGen         -- ^ what seed was used
+    , usedSize       :: Int            -- ^ what was the test size
+    , reason         :: String         -- ^ what was the reason
+    , labels         :: [(String,Int)] -- ^ labels and frequencies found during all successful tests
+    , output         :: String         -- ^ printed output
     }
-  | NoExpectedFailure               -- the expected failure did not happen
-    { labels      :: [(String,Int)] -- ^ labels and frequencies found during all successful tests
+  | NoExpectedFailure                  -- the expected failure did not happen
+    { numTests       :: Int            -- ^ number of tests performed
+    , labels         :: [(String,Int)] -- ^ labels and frequencies found during all successful tests
+    , output         :: String         -- ^ printed output
     }
  deriving ( Show, Read )
 
@@ -74,6 +83,7 @@ stdArgs = Args
   , maxSuccess = 100
   , maxDiscard = 500
   , maxSize    = 100
+  , chatty     = True
 -- noShrinking flag?
   }
 
@@ -91,36 +101,55 @@ quickCheckResult p = quickCheckWithResult stdArgs p
 
 -- | Tests a property, using test arguments, produces a test result, and prints the results to 'stdout'.
 quickCheckWithResult :: Testable prop => Args -> prop -> IO Result
-quickCheckWithResult args p =
-  do tm  <- newTerminal
-     rnd <- case replay args of
+quickCheckWithResult a p =
+  do tm <- if chatty a then newStdioTerminal else newNullTerminal
+     rnd <- case replay a of
               Nothing      -> newStdGen
               Just (rnd,_) -> return rnd
      test MkState{ terminal          = tm
-                 , maxSuccessTests   = maxSuccess args
-                 , maxDiscardedTests = maxDiscard args
-                 , computeSize       = case replay args of
-                                         Nothing    -> computeSize (maxSuccess args) (maxSize args)
+                 , maxSuccessTests   = maxSuccess a
+                 , maxDiscardedTests = maxDiscard a
+                 , computeSize       = case replay a of
+                                         Nothing    -> computeSize'
                                          Just (_,s) -> \_ _ -> s
                  , numSuccessTests   = 0
                  , numDiscardedTests = 0
                  , collected         = []
                  , expectedFailure   = False
                  , randomSeed        = rnd
-                 , isShrinking       = False
                  , numSuccessShrinks = 0
                  , numTryShrinks     = 0
                  } (unGen (property p))
-  where computeSize maxSuccess maxSize n d
+  where computeSize' n d
           -- e.g. with maxSuccess = 250, maxSize = 100, goes like this:
           -- 0, 1, 2, ..., 99, 0, 1, 2, ..., 99, 0, 2, 4, ..., 98.
-          | n `roundTo` maxSize + maxSize <= maxSuccess ||
-            n >= maxSuccess ||
-            maxSuccess `mod` maxSize == 0 = n `mod` maxSize + d `div` 10
+          | n `roundTo` maxSize a + maxSize a <= maxSuccess a ||
+            n >= maxSuccess a ||
+            maxSuccess a `mod` maxSize a == 0 = n `mod` maxSize a + d `div` 10
           | otherwise =
-            (n `mod` maxSize) * maxSize `div` (maxSuccess `mod` maxSize) + d `div` 10
+            (n `mod` maxSize a) * maxSize a `div` (maxSuccess a `mod` maxSize a) + d `div` 10
         n `roundTo` m = (n `div` m) * m
 
+-- | Tests a property and prints the results and all test cases generated to 'stdout'.
+-- This is just a convenience function that means the same as 'quickCheck' '.' 'verbose'.
+verboseCheck :: Testable prop => prop -> IO ()
+verboseCheck p = quickCheck (verbose p)
+
+-- | Tests a property, using test arguments, and prints the results and all test cases generated to 'stdout'.
+-- This is just a convenience function that combines 'quickCheckWith' and 'verbose'.
+verboseCheckWith :: Testable prop => Args -> prop -> IO ()
+verboseCheckWith args p = quickCheckWith args (verbose p)
+
+-- | Tests a property, produces a test result, and prints the results and all test cases generated to 'stdout'.
+-- This is just a convenience function that combines 'quickCheckResult' and 'verbose'.
+verboseCheckResult :: Testable prop => prop -> IO Result
+verboseCheckResult p = quickCheckResult (verbose p)
+
+-- | Tests a property, using test arguments, produces a test result, and prints the results and all test cases generated to 'stdout'.
+-- This is just a convenience function that combines 'quickCheckWithResult' and 'verbose'.
+verboseCheckWithResult :: Testable prop => Args -> prop -> IO Result
+verboseCheckWithResult a p = quickCheckWithResult a (verbose p)
+
 --------------------------------------------------------------------------
 -- main test loop
 
@@ -131,7 +160,7 @@ test st f
   | otherwise                                    = runATest st f
 
 doneTesting :: State -> (StdGen -> Int -> Prop) -> IO Result
-doneTesting st f =
+doneTesting st _f =
   do -- CALLBACK done_testing?
      if expectedFailure st then
        putPart (terminal st)
@@ -147,13 +176,18 @@ doneTesting st f =
         ++ " tests (expected failure)"
          )
      success st
+     theOutput <- terminalOutput (terminal st)
      if expectedFailure st then
-       return Success{ labels = summary st }
+       return Success{ labels = summary st,
+                       numTests = numSuccessTests st,
+                       output = theOutput }
       else
-       return NoExpectedFailure{ labels = summary st }
+       return NoExpectedFailure{ labels = summary st,
+                                 numTests = numSuccessTests st,
+                                 output = theOutput }
   
 giveUp :: State -> (StdGen -> Int -> Prop) -> IO Result
-giveUp st f =
+giveUp st _f =
   do -- CALLBACK gave_up?
      putPart (terminal st)
        ( bold ("*** Gave up!")
@@ -162,8 +196,10 @@ giveUp st f =
       ++ " tests"
        )
      success st
+     theOutput <- terminalOutput (terminal st)
      return GaveUp{ numTests = numSuccessTests st
                   , labels   = summary st
+                  , output   = theOutput
                   }
 
 runATest :: State -> (StdGen -> Int -> Prop) -> IO Result
@@ -178,40 +214,39 @@ runATest st f =
        ++ ")"
         )
      let size = computeSize st (numSuccessTests st) (numDiscardedTests st)
-     MkRose mres ts <- protectRose (unProp (f rnd1 size))
-     res <- mres
+     MkRose res ts <- protectRose (reduceRose (unProp (f rnd1 size)))
      callbackPostTest st res
      
-     case ok res of
-       Just True -> -- successful test
+     case res of
+       MkResult{ok = Just True, stamp = stamp, expect = expect} -> -- successful test
          do test st{ numSuccessTests = numSuccessTests st + 1
                    , randomSeed      = rnd2
-                   , collected       = stamp res : collected st
-                   , expectedFailure = expect res
+                   , collected       = stamp : collected st
+                   , expectedFailure = expect
                    } f
        
-       Nothing -> -- discarded test
+       MkResult{ok = Nothing, expect = expect} -> -- discarded test
          do test st{ numDiscardedTests = numDiscardedTests st + 1
                    , randomSeed        = rnd2
-                   , expectedFailure   = expect res
+                   , expectedFailure   = expect
                    } f
          
-       Just False -> -- failed test
+       MkResult{ok = Just False} -> -- failed test
          do if expect res
               then putPart (terminal st) (bold "*** Failed! ")
               else putPart (terminal st) "+++ OK, failed as expected. "
-            putTemp (terminal st)
-              ( short 30 (P.reason res)
-             ++ " (after "
-             ++ number (numSuccessTests st+1) "test"
-             ++ ")..."
-              )
-            foundFailure st res ts
+            numShrinks <- foundFailure st res ts
+            theOutput <- terminalOutput (terminal st)
             if not (expect res) then
-              return Success{ labels = summary st }
+              return Success{ labels = summary st,
+                              numTests = numSuccessTests st+1,
+                              output = theOutput }
              else
               return Failure{ usedSeed    = randomSeed st -- correct! (this will be split first)
                             , usedSize    = size
+                            , numTests    = numSuccessTests st+1
+                            , numShrinks  = numShrinks
+                            , output      = theOutput
                             , reason      = P.reason res
                             , labels      = summary st
                             }
@@ -232,7 +267,7 @@ summary st = reverse
 
 success :: State -> IO ()
 success st =
-  case labels ++ covers of
+  case allLabels ++ covers of
     []    -> do putLine (terminal st) "."
     [pt]  -> do putLine (terminal st)
                   ( " ("
@@ -242,16 +277,16 @@ success st =
     cases -> do putLine (terminal st) ":"
                 sequence_ [ putLine (terminal st) pt | pt <- cases ]
  where
-  labels = reverse
-         . sort
-         . map (\ss -> (showP ((length ss * 100) `div` numSuccessTests st) ++ head ss))
-         . group
-         . sort
-         $ [ concat (intersperse ", " s')
-           | s <- collected st
-           , let s' = [ t | (t,0) <- s ]
-           , not (null s')
-           ]
+  allLabels = reverse
+            . sort
+            . map (\ss -> (showP ((length ss * 100) `div` numSuccessTests st) ++ head ss))
+            . group
+            . sort
+            $ [ concat (intersperse ", " s')
+              | s <- collected st
+              , let s' = [ t | (t,0) <- s ]
+              , not (null s')
+              ]
   
   covers = [ ("only " ++ show occurP ++ "% " ++ fst (head lps) ++ "; not " ++ show reqP ++ "%")
            | lps <- groupBy first
@@ -277,38 +312,45 @@ success st =
 --------------------------------------------------------------------------
 -- main shrinking loop
 
-foundFailure :: State -> P.Result -> [Rose (IO P.Result)] -> IO ()
+foundFailure :: State -> P.Result -> [Rose P.Result] -> IO Int
 foundFailure st res ts =
-  do localMin st{ numTryShrinks = 0, isShrinking = True } res ts
+  do localMin st{ numTryShrinks = 0 } res ts
 
-localMin :: State -> P.Result -> [Rose (IO P.Result)] -> IO ()
-localMin st res [] = localMinFound st res
+localMin :: State -> P.Result -> [Rose P.Result] -> IO Int
 localMin st res _ | P.interrupted res = localMinFound st res
+localMin st res ts = do
+  putTemp (terminal st)
+    ( short 26 (P.reason res)
+   ++ " (after " ++ number (numSuccessTests st+1) "test"
+   ++ concat [ " and "
+            ++ show (numSuccessShrinks st)
+            ++ concat [ "." ++ show (numTryShrinks st) | numTryShrinks st > 0 ]
+            ++ " shrink"
+            ++ (if numSuccessShrinks st == 1
+                && numTryShrinks st == 0
+                then "" else "s")
+             | numSuccessShrinks st > 0 || numTryShrinks st > 0
+             ]
+   ++ ")..."
+    )
+  r <- tryEvaluate ts
+  case r of
+    Left err ->
+      localMinFound st
+         (exception "Exception while generating shrink-list" err) { callbacks = callbacks res }
+    Right ts' -> localMin' st res ts'
 
-localMin st res (t : ts) =
+localMin' :: State -> P.Result -> [Rose P.Result] -> IO Int
+localMin' st res [] = localMinFound st res
+localMin' st res (t:ts) =
   do -- CALLBACK before_test
-     MkRose mres' ts' <- protectRose t
-     res' <- mres'
-     putTemp (terminal st)
-       ( short 35 (P.reason res)
-      ++ " (after " ++ number (numSuccessTests st+1) "test"
-      ++ concat [ " and "
-               ++ show (numSuccessShrinks st)
-               ++ concat [ "." ++ show (numTryShrinks st) | numTryShrinks st > 0 ]
-               ++ " shrink"
-               ++ (if numSuccessShrinks st == 1
-                   && numTryShrinks st == 0
-                   then "" else "s")
-                | numSuccessShrinks st > 0 || numTryShrinks st > 0
-                ]
-      ++ ")..."
-       )
-     callbackPostTest st res'
-     if ok res' == Just False
-       then foundFailure st{ numSuccessShrinks = numSuccessShrinks st + 1 } res' ts'
-       else localMin st{ numTryShrinks = numTryShrinks st + 1 } res ts
+    MkRose res' ts' <- protectRose (reduceRose t)
+    callbackPostTest st res'
+    if ok res' == Just False
+      then foundFailure st{ numSuccessShrinks = numSuccessShrinks st + 1 } res' ts'
+      else localMin st{ numTryShrinks = numTryShrinks st + 1 } res ts
 
-localMinFound :: State -> P.Result -> IO ()
+localMinFound :: State -> P.Result -> IO Int
 localMinFound st res =
   do putLine (terminal st)
        ( P.reason res
@@ -319,17 +361,18 @@ localMinFound st res =
       ++ "):  "
        )
      callbackPostFinalFailure st res
+     return (numSuccessShrinks st)
 
 --------------------------------------------------------------------------
 -- callbacks
 
 callbackPostTest :: State -> P.Result -> IO ()
 callbackPostTest st res =
-  sequence_ [ f st res | PostTest f <- callbacks res ]
+  sequence_ [ f st res | PostTest _ f <- callbacks res ]
 
 callbackPostFinalFailure :: State -> P.Result -> IO ()
 callbackPostFinalFailure st res =
-  sequence_ [ f st res | PostFinalFailure f <- callbacks res ]
+  sequence_ [ f st res | PostFinalFailure _ f <- callbacks res ]
 
 --------------------------------------------------------------------------
 -- the end.
